/**
 * This file is part of Princess, a theorem prover for Presburger
 * arithmetic with uninterpreted predicates.
 * <http://www.philipp.ruemmer.org/princess.shtml>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * Princess is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Princess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Princess.  If not, see <http://www.gnu.org/licenses/>.
 */

package ap.parser

import ap.terfor.{ConstantTerm, TermOrder}
import ap.parameters.{PreprocessingSettings, Param}

/**
 * Preprocess an InputAbsy formula in order to make it suitable for
 * proving. The result is a list of formulae, because the original formula
 * may contain named parts (<code>INamedPart</code>).
 */
object Preprocessing {
  
  class PreprocessingException(msg : String) extends Exception(msg)

  private def withinParts(f : IFormula, fun : (IFormula) => IFormula) = f match {
      case INamedPart(name, subFor) =>
        f update List(fun(subFor))
      case _ =>
        fun(f)
    }
  
  def apply(f : IFormula,
            interpolantSpecs : List[IInterpolantSpec],
            signature : Signature,
            settings : PreprocessingSettings)
            : (List[IFormula], List[IInterpolantSpec], Signature) = {
              
    // turn the formula into a list of its named parts
    val fors = PartExtractor(f)
    
    // expand equivalences
    val fors2 = for (f <- fors) yield EquivExpander(f)
    
    // translate functions to relations
    val functionEncoder = new FunctionEncoder
    var order3 = signature.order
    val fors3 = for (f <- fors2) yield withinParts(f, (subFor) => { 
      val (g, o) = functionEncoder(subFor, order3)
      order3 = o
      g
    })
    val fors4 = functionEncoder.axioms match {
      case IBoolLit(true) => fors3
      case x => !x :: fors3
    }
    
    // do clausification
    val fors5 = Param.CLAUSIFIER(settings) match {
      case Param.ClausifierOptions.None =>
        fors4
      case Param.ClausifierOptions.Simple =>
        for (f <- fors4) yield SimpleClausifier(f)
    }
    
    (fors5, interpolantSpecs, signature updateOrder order3)
  }
  
}
